using System;
using System.Collections.Generic;
using System.Text;
//
using System.Runtime.InteropServices;
using System.IO;
using System.Net;
using System.Net.Mail;
using System.Net.Mime;
using System.Threading;

namespace UniPaaSUtils {

    #region Internal Helper Classes
    /// <summary>
    /// Format melhor os dados de uma 'exceo' ocorrida
    /// </summary>
    internal abstract class ExceptionAnalyzer {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="message"></param>
        /// <param name="e"></param>
        /// <returns></returns>
        public static string MsgBuilder(string message, Exception e) {
            string result = message;
            string st = string.Empty;
            while (e != null) {
                st = string.IsNullOrEmpty(st) ? "\r\n" + e.StackTrace : st;
                result += " -> " + e.Message + " [from " + e.Source + "]";
                e = e.InnerException;
            }
            return (result + st);
        }
    }
    #endregion

    #region SMTP Email
    internal enum RetCodes : int {
        R_NOERROR = 0,
        R_RUNTIME_ERR = -1,
        R_SMTP_NO_SERVER = -2,
        R_SMTP_NO_FROM = -3,
        R_SMTP_NO_TO = -4,
        R_UNKNOWN_ERR = -5
    }

    /// <summary>
    /// Interface da classe
    /// </summary>
    [Guid(@"8941A2BF-3976-4D34-9FD9-A2FCD24BF74A")]
    [InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface IMailSmtp {
        [DispId(1)]
        bool SendMessage(string server,
                         int port,
                         int tOut,
                         string user,
                         string pwd,
                         bool ssl,
                         string from,
                         string ToList,
                         string CCList,
                         string BCCList,
                         string Subject,
                         string Body,
                         string AttachmentList,
                         ref int ReturnCode,
                         ref string ReturnText);
    }

    /// <summary>
    /// Implementao fsica da classe
    /// </summary>
    [Guid(@"8822F9B9-261B-4A69-AE25-423BF6BF6936")]
    [ClassInterface(ClassInterfaceType.None)]
    public class MailSmtp : IMailSmtp {

        private bool sending = false;
        private bool cancelled = false;
        private SmtpException sendErrorSmtp = null;

        /// <summary>
        /// Transmisso assncrona do e-mail foi co0mpletada, seja qual for o motivo
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void smtp_SendCompleted(object sender, System.ComponentModel.AsyncCompletedEventArgs e) {
            if (e.Error != null) {
                // Erro no envio de e-mail
                string errMsg = e.Error.Message;
                Exception stack = e.Error.InnerException;
                while (stack != null) {
                    errMsg += " -> " + stack.Message;
                    stack = stack.InnerException;
                }
                sendErrorSmtp = new SmtpException(SmtpStatusCode.TransactionFailed, errMsg);
            } else if (e.Cancelled) {
                // Cancelado por Time-Out
                cancelled = true;
                sendErrorSmtp = new SmtpException(SmtpStatusCode.GeneralFailure, string.Format("Operation timed out: {0} seconds", (int)e.UserState));
            }
            sending = false;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="server">[Obrigatrio] Servidor SMTP para conexo</param>
        /// <param name="port">Porta de conexo neste servidor (default: 25)</param>
        /// <param name="tOut">TimeOut em milisegundos para a operao (default: 30seg)</param>
        /// <param name="user">Usurio (se necessrio) para login no servidor</param>
        /// <param name="pwd">Senha (se necessrio) para login no servidor</param>
        /// <param name="ssl">Usar conexo segura (SSL/TLS)? (default: false)</param>
        /// <param name="from">[Obrigatrio] Remetente</param>
        /// <param name="ToList">[Obrigatrio se 'CC' e 'BCC' vazios] Lista (separada por ',' qdo + de 1) dos destinatrios</param>
        /// <param name="CCList">[Obrigatrio se 'TO' e 'BCC' vazios] Lista (separada por ',' qdo + de 1) dos destinatrios em cpia</param>
        /// <param name="BCCList">[Obrigatrio se 'TO' e 'CC' vazios] Lista (separada por ',' qdo + de 1) dos destinatrios em cpia oculta</param>
        /// <param name="Subject">Assunto da mensagem</param>
        /// <param name="Body">Contedo da mensagem</param>
        /// <param name="AttachmentList">Lista (separada por ',' qdo + de 1) dos nomes completos dos arquivos a anexar  mensagem</param>
        /// <param name="ReturnCode">Ret: 0 (sucesso), <> 0 (erro)</param>
        /// <param name="ReturnText">Ret: se sucesso: "", se erro: descrio do erro</param>
        /// <returns></returns>
        public bool SendMessage(string server,
                                int port,
                                int tOut,
                                string user,
                                string pwd,
                                bool ssl,
                                string from,
                                string ToList,
                                string CCList,
                                string BCCList,
                                string Subject,
                                string Body,
                                string AttachmentList,
                                ref int ReturnCode,
                                ref string ReturnText) {
            //
            ReturnCode = (int)RetCodes.R_UNKNOWN_ERR;
            ReturnText = @"Unknown Error";
            //
            try {
                if (string.IsNullOrEmpty(server)) {
                    // o argumento 'server'  obrigatrio
                    ReturnCode = (int)RetCodes.R_SMTP_NO_SERVER;
                    ReturnText = @"No server information to send the message";
                } else if (string.IsNullOrEmpty(from)) {
                    // o argumento 'from'  obrigatrio
                    ReturnCode = (int)RetCodes.R_SMTP_NO_FROM;
                    ReturnText = @"No sender (FROM) information to send the message";
                } else {
                    port = (port <= 0) || (port >= 65535) ? 25 : port;
                    SmtpClient smtp = new SmtpClient(server, port);
                    if (!string.IsNullOrEmpty(user)) {
                        // Precisa autenticar no servidor SMTP
                        NetworkCredential loginInfo = new NetworkCredential(user, pwd);
                        smtp.UseDefaultCredentials = false;
                        smtp.Credentials = loginInfo;
                    }
                    MailAddress mailFrom = new MailAddress(from, from, Encoding.UTF8);
                    MailMessage mailMsg = null;
                    // (to)
                    if (!string.IsNullOrEmpty(ToList)) {
                        // Destinatrios
                        string[] to = ToList.Split(new char[] { ',' });
                        foreach (string mailTo in to) {
                            MailAddress mailAddr = new MailAddress(mailTo, mailTo, Encoding.UTF8);
                            if (mailMsg == null) {
                                mailMsg = new MailMessage(mailFrom, mailAddr);
                            } else {
                                mailMsg.To.Add(mailAddr);
                            }
                        }
                    }
                    // (cc)
                    if (!string.IsNullOrEmpty(CCList)) {
                        // Em cpia
                        string[] cc = CCList.Split(new char[] { ',' });
                        foreach (string mailCC in cc) {
                            MailAddress mailAddr = new MailAddress(mailCC, mailCC, Encoding.UTF8);
                            if (mailMsg == null) {
                                mailMsg = new MailMessage(mailFrom, mailAddr);
                            } else {
                                mailMsg.To.Add(mailAddr);
                            }
                        }
                    }
                    // (bcc)
                    if (!string.IsNullOrEmpty(BCCList)) {
                        // Em cpia oculta
                        string[] bcc = BCCList.Split(new char[] { ',' });
                        foreach (string mailBCC in bcc) {
                            MailAddress mailAddr = new MailAddress(mailBCC, mailBCC, Encoding.UTF8);
                            if (mailMsg == null) {
                                mailMsg = new MailMessage(mailFrom, mailAddr);
                            } else {
                                mailMsg.To.Add(mailAddr);
                            }
                        }
                    }
                    //
                    if (mailMsg == null) {
                        // nenhum TO, CC ou BCC
                        ReturnCode = (int)RetCodes.R_SMTP_NO_TO;
                        ReturnText = @"No destinatary (TO, CC or BCC) information to send the message";
                    } else {
                        Subject = string.IsNullOrEmpty(Subject) ? string.Empty : Subject;
                        Body = string.IsNullOrEmpty(Body) ? string.Empty : Body;
                        tOut = tOut <= 0 ? 30000 : tOut;
                        // subject
                        mailMsg.SubjectEncoding = Encoding.UTF8;
                        mailMsg.Subject = Subject;
                        // message
                        mailMsg.BodyEncoding = Encoding.UTF8;
                        mailMsg.Body = Body;
                        // (attachments)
                        if (!string.IsNullOrEmpty(AttachmentList)) {
                            // incluir anexos na mensagem
                            string[] fileNames = AttachmentList.Split(new char[] { ',' });
                            foreach (string fileName in fileNames) {
                                if (File.Exists(fileName.Trim())) {
                                    Attachment file = new Attachment(fileName.Trim(), MediaTypeNames.Application.Octet);
                                    mailMsg.Attachments.Add(file);
                                }
                            }
                        }
                        //
                        // Send the Email
                        //
                        smtp.EnableSsl = ssl;
                        smtp.Timeout = tOut;
                        smtp.SendCompleted += new SendCompletedEventHandler(smtp_SendCompleted);
                        sending = true;
                        cancelled = false;
                        sendErrorSmtp = null;
                        smtp.SendAsync(mailMsg, tOut / 1000); // inicia o envio assncrono
                        //
                        // Aguarda a finalizao do envio, at o limite de tempo definido em 'tOut'
                        //
                        DateTime dtLimit = DateTime.Now.AddMilliseconds(tOut);
                        while (sending) {
                            Thread.Sleep(10); // refresh do processador
                            if ((!cancelled) && (DateTime.Now > dtLimit)) {
                                // cancela o envio do e-mail
                                cancelled = true;
                                smtp.SendAsyncCancel();
                            }
                        }
                        //
                        // Free Resources
                        //
                        foreach (Attachment file in mailMsg.Attachments) {
                            file.Dispose();
                        }
                        mailMsg.Dispose();
                        //
                        if (sendErrorSmtp != null) {
                            // Erro no e-mail
                            throw sendErrorSmtp;
                        }
                        //
                        ReturnCode = (int)RetCodes.R_NOERROR;
                        ReturnText = string.Empty;
                    }
                }

            } catch (SmtpFailedRecipientsException ex) {
                ReturnCode = Math.Abs((int)ex.StatusCode);
                ReturnText = ExceptionAnalyzer.MsgBuilder(string.Format(@"(Email Send) Recipients Failure: {0} {1} -> {2}", (int)ex.StatusCode, ex.StatusCode, ex.Message), ex.InnerException);
            } catch (SmtpException me) {
                ReturnCode = Math.Abs((int)me.StatusCode);
                ReturnText = ExceptionAnalyzer.MsgBuilder(string.Format(@"(Email Send) Transmission Failure: {0} {1} -> {2}", (int)me.StatusCode, me.StatusCode, me.Message), me.InnerException);
            } catch (Exception e) {
                ReturnCode = (int)RetCodes.R_RUNTIME_ERR;
                ReturnText = ExceptionAnalyzer.MsgBuilder("(Email Send)", e);
            }
            //
            return (ReturnCode == (int)RetCodes.R_NOERROR);
        }
    }
    #endregion
}


// EOF